% MATLAB code:  Multivariate Regression (Chapter 11, Miscellanea)
% File: mlts.m
% Coded by: Wai-Sum Chan, et al. (2015)
%
% INPUT:
%     x: data-matrix (n,p)
%     y: data-matrix (n,q)
%     gamma: proportion of trimming 
%     options: (field)
%       - options.ns :    contains number of subsets 
%       - options.nc :    number os C-steps 
%       - options.delta : critical value for Reweighted estimator
% OUTPUT:
%     beta:   matrix (p,q) of MLTS-regression coefficients
%     sigma:  matrix (q,q) containing MLTS-residual covariance
%     dres:   residual distances (n,1) w.r.t. initial fit
%     betaR:  matrix (p,q) of RMLTS-regression coefficients
%     sigmaR: matrix (q,q) containing RMLTS-residual covariance
%     dresR:  residual distances (n,1) w.r.t. RMLTS
%
% Remark: if intercept needed, add a column of ones to the x-matrix.
%  
% Example:
%    x      = [ones(50,1), randn(50,2)];
%    beta   = ones(3,2);
%    y      = x*beta+randn(50,2);
%    output = mlts(x,y,0.25);
%
% Reference:
% Chan, W.S., Cheung, S.H., Chow, W.K., and Zhang, L-X. (2015).
%   https://doi.org/10.1002/for.2344
%   time series analysis.
%   Journal of Forecasting, 34(6), 441-454.
%   DOI: 10.1002/for.2344.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function result = mlts(x,y,gamma,options)

if nargin < 4
  ns = 500; nc = 10; delta = 0.01;
else
    ns=options.ns; nc=options.nc; delta=options.delta;
end

p = size(x,2); 
q = size(y,2); 
n = size(x,1);
h = floor(n*(1-gamma))+1;

if h>n
  h = n;
else
  h = h;
end

obj0 = 1e10;
for i=1:ns
    [us,istart] = sort(rand(n,1));        % us stores the random numbers, istart stores the rank;
    istart      = istart(1:(p+q));        % Return the first (p+q) rows of istart;
    xstart      = x(istart,:);            % Return the corresponding subset of X;
    ystart      = y(istart,:);            % Return the corresponding subset of Y;
    bstart      = xstart\ystart;          % Compute the corresponding Beta of the subset of X and Y;
    sigmastart  = (ystart-xstart*bstart)'*(ystart-xstart*bstart)/q;   % Compute the 
                                          % corresponding Variance of the subset of X and Y;
                                     
    for j=1:nc
        res   = y-x*bstart;          % Compute the residuals of the particular model above;
        dist2 = sum((res*inv(sigmastart)).*res,2); % Compute the row sum of the residual Mahalanobis Distances;
        [dist2s,idist2] = sort(dist2); % dist2s stores the Mahalanobis distances, idist2 stores the ranks;
        idist2 = idist2(1:h);        % Return the corresponding subset;
        xstart = x(idist2,:);
        ystart = y(idist2,:);        % Return the subsets of X and Y;
        bstart = inv(xstart'*xstart)*xstart'*ystart;    % Compute the Beta;
        sigmastart = (ystart-xstart*bstart)'*(ystart-xstart*bstart)/(h-p);  % Compute 
                                                                   % the corresponding Variance;
    end
    obj = det(sigmastart);           % Compute the determinant of the estimated variance;
    if obj < obj0
        result.beta  = bstart;
        result.sigma = sigmastart;
        obj0         = obj;
    end
end

cgamma       = (1-gamma)/chi2cdf(chi2inv(1-gamma,q),q+2);
result.sigma = cgamma*result.sigma;
res          = y-x*result.beta;
result.dres  = sum((res*inv(result.sigma)).*res,2);
result.dres  = (result.dres).^(0.5);             % The Mahalanobis distances;


qdelta    = sqrt(chi2inv(1-delta,q));
nooutlier = (result.dres<=qdelta*ones(n,1));     % Compute which components should be 
                                                 % trimmed by RMLTS;
xgood     = x(nooutlier,:);ygood=y(nooutlier,:); % Trimmed out the corresponding components 
                                                 % by RMLTS;
result.betaR   = inv(xgood'*xgood)*xgood'*ygood; % Compute the Beta of X and Y by RMLTS;
result.sigmaR  = (ygood-xgood*result.betaR)'*(ygood-xgood*result.betaR)/(sum(nooutlier)-p); 
                                                 % Compute Variance;
cdelta           = (1-delta)/chi2cdf(qdelta^2,q+2); 
result.sigmaR    = cdelta*result.sigmaR;
resR             = y-x*result.betaR;
result.dresR     = sum((resR*inv(result.sigmaR)).*resR,2);
result.dresR     = (result.dresR).^(0.5);
result.nooutlier = nooutlier;
result.cdelta    = cdelta;
result.cgamma    = cgamma;
result.qdelta    = qdelta;